﻿using Priority_Queue;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using TemporalSourceCover.Utility;

namespace TemporalSourceCover
{
    public sealed class TemporalGraph
    {

        private Interval _lifetime = new Interval(Time.MaxValue, Time.MinValue);

        /// <summary>
        /// Lifetime of the temporal graph. Does not change if vertices are removed!
        /// </summary>
        public Interval Lifetime => _lifetime;


        private readonly Dictionary<long, TemporalGraphVertex> _vertices = new();
        /// <summary>
        /// Contains the vertices accessible by their id
        /// </summary>
        public IReadOnlyDictionary<long, TemporalGraphVertex> Vertices => _vertices;

        /// <summary>
        /// Returns the vertex with the provided index if it exists, otherwise null.
        /// </summary>
        /// <param name="id"></param>
        /// <returns></returns>
        public TemporalGraphVertex? this[long id] => _vertices.TryGetValue(id, out var vertex) ? vertex : default;

        private readonly SortedSet<TemporalGraphEdge> _edges = new(new TemporalGraphEdgeComparer());
        /// <summary>
        /// Contains all edges
        /// </summary>
        public IReadOnlySet<TemporalGraphEdge> Edges => _edges;


        private readonly Dictionary<VertexType, HashSet<TemporalGraphVertex>> _verticesByType = Enum.GetValues<VertexType>().ToDictionary(t => t, _ => new HashSet<TemporalGraphVertex>());
        /// <summary>
        /// Contains a set of vertices grouped by their current type
        /// </summary>
        public readonly IReadOnlyDictionary<VertexType, IReadOnlySet<TemporalGraphVertex>> VerticesByType; // no property as we don't want to create a new dictionary every time we access this field/property

        private readonly HashSet<TemporalGraphVertex> _blackVertices = new();
        /// <summary>
        /// Contains all black vertices
        /// </summary>
        public IReadOnlySet<TemporalGraphVertex> BlackVertices => _blackVertices;

        private readonly HashSet<TemporalGraphVertex> _redVertices = new();
        /// <summary>
        /// Contains all red vertices
        /// </summary>
        public IReadOnlySet<TemporalGraphVertex> RedVertices => _redVertices;

        internal TemporalGraphTransaction? CurrentTransaction { get; set; }

        public TemporalGraph()
        {
            VerticesByType = _verticesByType.ToDictionary(t => t.Key, t => (IReadOnlySet<TemporalGraphVertex>)t.Value);
        }


        internal bool NoUpdate { get; private set; } = false;
        private class DisposedUpdater : IDisposable
        {
            private readonly TemporalGraph _graph;

            public DisposedUpdater(TemporalGraph graph)
            {
                _graph = graph;
                _graph.NoUpdate = true;
            }

            public void Dispose()
            {
                _graph.NoUpdate = false;

                foreach (var vertex in _graph._vertices.Values)
                {
                    vertex.UpdateType();
                }
            }
        }

        public IDisposable DontUpdate()
        {
            return new DisposedUpdater(this);
        }


        public void AddEdge(long u, long v, Time time, bool undirected = false)
        {
            // skip edges to itself
            if (u == v)
                return;

            var from = GetOrAddVertex(u);
            var to = GetOrAddVertex(v);

            AddEdge(from, to, time, undirected);

            _lifetime.Include(time);
        }

        public TemporalGraph Clone()
        {
            var clone = new TemporalGraph();

            using (clone.DontUpdate())
            {
                foreach (var vertex in _vertices.Values)
                {
                    foreach (var (u, v, t) in vertex.OutEdges)
                    {
                        clone.AddEdge(u.Id, v.Id, t);
                    }
                }
            }

            return clone;
        }

        internal void AddEdge(TemporalGraphEdge edge)
        {
            if (!_edges.Add(edge))
                throw new InvalidOperationException();

            edge.From.AddEdge(edge);
            edge.To.AddEdge(edge);
        }

        public void AddEdge(TemporalGraphVertex from, TemporalGraphVertex to, Time time, bool undirected = false)
        {
            if (from == to)
                throw new InvalidOperationException();

            var edge = new TemporalGraphEdge(from, to, time);

            if (from.AddEdge(edge) & to.AddEdge(edge))
            {
                _edges.Add(edge);
            }

            if (undirected)
            {
                var backEdge = new TemporalGraphEdge(to, from, time);

                if (from.AddEdge(backEdge) & to.AddEdge(backEdge))
                {
                    _edges.Add(backEdge);
                }
            }
        }

        public void RemoveEdge(TemporalGraphEdge edge)
        {
            Timeout.ThrowOnCancel();

            var (from, to, _) = edge;

            from.RemoveEdge(edge);
            to.RemoveEdge(edge);

            _edges.Remove(edge);

            CurrentTransaction?.RemovedEdges.Add(edge);
        }

        /// <summary>
        /// Removes the provided source vertex and all its traversable edges from the graph. 
        /// Additionally sets the time of all reached vertices 
        /// Returns a set containing the reached vertices.
        /// Returns a set containing all removed edges.
        /// </summary>
        /// <param name="source"></param>
        public void RemoveSourceAndPaths(TemporalGraphVertex source, out HashSet<TemporalGraphVertex> reachedVertices, out HashSet<TemporalGraphEdge> removedEdges)
        {
            var reached = GetReachableVertices(source, out var reachedEdges);
            foreach (var (vertex, time) in reached)
            {
                vertex.ReachedAt = Time.Min(vertex.ReachedAt ?? Time.MaxValue, time);
            }

            if (reachedEdges.Count > 0)
            {
                using (DontUpdate())
                {
                    foreach (var removeEdge in reachedEdges)
                    {
                        RemoveEdge(removeEdge);
                    }
                }
            }

            removedEdges = reachedEdges;
            reachedVertices = reached.Keys.ToHashSet();
        }


        /// <summary>
        /// Removes a vertex and all adjacent edges from the graph.
        /// </summary>
        /// <param name="vertex"></param>
        /// <returns>Number of removed edges</returns>
        public int RemoveVertex(TemporalGraphVertex vertex)
        {
            Timeout.ThrowOnCancel();

            _vertices.Remove(vertex.Id);
            CurrentTransaction?.RemovedVertices.Add(vertex);

            _verticesByType[vertex.Type].Remove(vertex);
            if (vertex.IsRed)
            {
                _redVertices.Remove(vertex);
            }
            else
            {
                _blackVertices.Remove(vertex);
            }

            vertex.VertexTypeChanged -= OnVertexTypeChanged;
            vertex.VertexReachedAtTimeChanged -= OnVertexReachedAtTimeChanged;


            int removedEdges = vertex.InEdges.Count + vertex.OutEdges.Count;

            while (vertex.InEdges.Count > 0)
            {
                var edge = vertex.InEdges.First();
                RemoveEdge(edge);
            }

            while (vertex.OutEdges.Count > 0)
            {
                var edge = vertex.OutEdges.First();
                RemoveEdge(edge);
            }

            return removedEdges;
        }

        internal void AddVertex(TemporalGraphVertex vertex)
        {
            if (_vertices.ContainsKey(vertex.Id))
                throw new InvalidOperationException();
            _vertices.Add(vertex.Id, vertex);

            if (!_verticesByType[vertex.Type].Add(vertex))
                throw new InvalidOperationException();

            if (vertex.IsRed && !_redVertices.Add(vertex))
                throw new InvalidOperationException();
            else if (!vertex.IsRed && !_blackVertices.Add(vertex))
                throw new InvalidOperationException();


            vertex.VertexTypeChanged += OnVertexTypeChanged;
            vertex.VertexReachedAtTimeChanged += OnVertexReachedAtTimeChanged;
        }

        private TemporalGraphVertex GetOrAddVertex(long id)
        {
            if (_vertices.TryGetValue(id, out var vertex))
                return vertex;

            vertex = new TemporalGraphVertex(id, this);
            _vertices.Add(id, vertex);

            _verticesByType[vertex.Type].Add(vertex);
            _blackVertices.Add(vertex);


            vertex.VertexTypeChanged += OnVertexTypeChanged;
            vertex.VertexReachedAtTimeChanged += OnVertexReachedAtTimeChanged;

            return vertex;
        }

        private void OnVertexReachedAtTimeChanged(TemporalGraphVertex vertex, Time? before, Time? after)
        {
            // only track first change in time
            if (CurrentTransaction?.OldReachedTimes.ContainsKey(vertex) == false)
            {
                CurrentTransaction.OldReachedTimes.Add(vertex, before);
            }

            if (before is null && after is not null)
            {
                _blackVertices.Remove(vertex);
                _redVertices.Add(vertex);
            }
            else if (before is not null && after is null)
            {
                _redVertices.Remove(vertex);
                _blackVertices.Add(vertex);
            }
        }

        private void OnVertexTypeChanged(TemporalGraphVertex vertex, VertexType oldType, VertexType newType)
        {
            _verticesByType[oldType].Remove(vertex);
            _verticesByType[newType].Add(vertex);
        }

        private sealed class TemporalGraphVertexPriorityNode : FastPriorityQueueNode
        {
            public TemporalGraphVertex Vertex { get; }
            public Time Time { get; set; }

            public TemporalGraphVertexPriorityNode(TemporalGraphVertex vertex) : this(vertex, default)
            {

            }

            public TemporalGraphVertexPriorityNode(TemporalGraphVertex vertex, Time time)
            {
                Vertex = vertex;
                Time = time;
            }

            public void Deconstruct(out TemporalGraphVertex vertex, out Time time)
            {
                vertex = Vertex;
                time = Time;
            }
        }

        public Dictionary<TemporalGraphVertex, Time> GetReachableVertices(TemporalGraphVertex source, out HashSet<TemporalGraphEdge> reachedEdges)
        {
            var reachedNodes = new Dictionary<TemporalGraphVertex, Time>();
            reachedEdges = new HashSet<TemporalGraphEdge>();

            var qNodes = new Dictionary<TemporalGraphVertex, TemporalGraphVertexPriorityNode>()
            {
                { source, new TemporalGraphVertexPriorityNode(source, Time.MinValue) }
            };
            var q = new FastPriorityQueue<TemporalGraphVertexPriorityNode>(Vertices.Count);
            q.Enqueue(qNodes[source], Time.MinValue.Value);

            var lastTime = Time.MinValue;

            while (q.Count > 0)
            {
                var (node, time) = q.Dequeue();
                reachedNodes.Add(node, time);

                if (lastTime > time)
                {
                    throw new InvalidOperationException();
                }
                lastTime = time;

                foreach (var (to, edges) in node.OutNeighbors)
                {
                    if (edges.Count == 0)
                        throw new InvalidOperationException();

                    var sourceReachedEdges = edges.SkipWhile(e => !Time.Proceeds(time, e.Time)).ToList();
                    if (sourceReachedEdges.Count == 0)
                        continue;

                    var edge = sourceReachedEdges[0];
                    if (!reachedNodes.TryGetValue(to, out var toTime))
                    {
                        if (!qNodes.TryGetValue(to, out var prioTo))
                        {
                            prioTo = new TemporalGraphVertexPriorityNode(to, edge.Time);
                            qNodes.Add(to, prioTo);
                            if (to.Id == 15589)
                            {

                            }
                            q.Enqueue(prioTo, edge.Time.Value);
                        }
                        else if (prioTo.Time > edge.Time)
                        {
                            prioTo.Time = edge.Time;
                            q.UpdatePriority(prioTo, edge.Time.Value);
                        }
                    }

                    reachedEdges.AddRange(sourceReachedEdges);
                }
            }

            return reachedNodes;
        }

        public IReadOnlySet<TemporalGraphVertex> GetReachableVertices(TemporalGraphVertex source) => GetReachableVertices(source, out _).Keys.ToHashSet();

        private TemporalGraphReachability GetFullReachability_Parallel()
        {
            using (Metrics.BeginMeasurement("Full Reachability", log: false))
            {
                var reachabilityBySource = _vertices.Values.ToDictionary(v => v, v => new HashSet<TemporalGraphVertex>()); // reachabilityBySource[v] contains all vertices that can be reached by v
                var reachabilityByTarget = _vertices.Values.ToDictionary(v => v, v => new HashSet<TemporalGraphVertex>()); // reachabilityByTarget[v] contains all vertices that reach v

                Parallel.ForEach(_vertices.Values, vertex =>
                {
                    Timeout.ThrowOnCancel();
                    var targets = GetReachableVertices(vertex);

                    reachabilityBySource[vertex].AddRange(targets);

                    foreach (var target in targets)
                    {
                        Timeout.ThrowOnCancel();
                        var targetReach = reachabilityByTarget[target];
                        lock (targetReach)
                        {
                            targetReach.Add(vertex);
                        }
                    }
                });
                return new TemporalGraphReachability(reachabilityBySource, reachabilityByTarget);
            }
        }

        public TemporalGraphReachability GetFullReachability()
        {
            return GetFullReachability_Parallel();
        }
    }
}
