﻿using System.Collections.Generic;

namespace TemporalSourceCover
{
    public sealed class TemporalGraphReachability
    {
        internal TemporalGraphReachabilityTransaction? CurrentTransaction { get; set; }

        public Dictionary<TemporalGraphVertex, HashSet<TemporalGraphVertex>> BySource { get; }
        public Dictionary<TemporalGraphVertex, HashSet<TemporalGraphVertex>> ByTarget { get; }

        public TemporalGraphReachability(
            Dictionary<TemporalGraphVertex, HashSet<TemporalGraphVertex>> bySource,
            Dictionary<TemporalGraphVertex, HashSet<TemporalGraphVertex>> byTarget)
        {
            BySource = bySource;
            ByTarget = byTarget;
        }

        public bool RemoveSource(TemporalGraphVertex source)
        {
            if (!BySource.TryGetValue(source, out var targets))
                return false;

            BySource.Remove(source);
            CurrentTransaction?.AddRemovedBySource(source, targets);

            foreach (var target in targets)
            {
                // target already ignored(covered by superset and removed from targets
                if (!ByTarget.TryGetValue(target, out var sources))
                    continue;

                if (sources.Remove(source))
                {
                    CurrentTransaction?.AddRemovedByTarget(target, source);
                }

                if (sources.Count == 0)
                {
                    RemoveTarget(target);
                }
            }

            return true;
        }

        public bool RemoveOnlyTarget(TemporalGraphVertex target)
        {
            if (!ByTarget.TryGetValue(target, out var sources))
                return false;

            ByTarget.Remove(target);
            CurrentTransaction?.AddRemovedByTarget(target, sources);
            return true;
        }

        public bool RemoveTarget(TemporalGraphVertex target)
        {
            if (!ByTarget.TryGetValue(target, out var sources))
                return false;

            ByTarget.Remove(target);
            CurrentTransaction?.AddRemovedByTarget(target, sources);

            foreach (var source in sources)
            {
                var sourceTargets = BySource[source];
                if (sourceTargets.Remove(target))
                {
                    CurrentTransaction?.AddRemovedBySource(source, target);
                }

                if (sourceTargets.Count == 0)
                {
                    RemoveSource(source);
                }
            }

            return true;
        }

        public TemporalGraphReachability Clone()
        {
            var clone = new TemporalGraphReachability(new(), new());

            foreach (var (source, targets) in BySource)
            {
                clone.BySource.Add(source, new HashSet<TemporalGraphVertex>(targets));
            }

            foreach (var (target, sources) in ByTarget)
            {
                clone.ByTarget.Add(target, new HashSet<TemporalGraphVertex>(sources));
            }

            return clone;
        }
    }
}
