﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;

namespace TemporalSourceCover
{
    public delegate void VertexTypeChangedHandler(TemporalGraphVertex vertex, VertexType oldType, VertexType newType);
    public delegate void VertexReachedAtTimeChangedHandler(TemporalGraphVertex vertex, Time? before, Time? after);

    [DebuggerDisplay("{_debug}")]
    public sealed class TemporalGraphVertex
    {
        private string _debug => $"{(Id <= 26 ? ((char)('a' + Id)).ToString() : "") + " "}Vertex {Id} (Reached: {IsRed}, Type: {Type}, in-edges: {InEdges.Count}, out-edges: {OutEdges.Count}, in-neighbors: {InNeighbors.Count}, out-neighbors: {OutNeighbors.Count})";

        public VertexType Type { get; private set; } = VertexType.Isolated; // a vertex has no edges initially
        internal event VertexTypeChangedHandler? VertexTypeChanged;
        internal event VertexReachedAtTimeChangedHandler? VertexReachedAtTimeChanged;

        private Time? _reachedAt;
        public Time? ReachedAt
        {
            get => _reachedAt;
            set
            {
                var oldTime = _reachedAt;
                _reachedAt = value;

                VertexReachedAtTimeChanged?.Invoke(this, oldTime, value);
            }
        }
        public bool IsRed => ReachedAt is not null;

        public readonly long Id;
        private readonly TemporalGraph _graph;

        /// <summary>
        /// Contains all edges from the in-neighbor
        /// </summary>
        public readonly Dictionary<TemporalGraphVertex, SortedSet<TemporalGraphEdge>> InNeighbors = new();

        public readonly SortedSet<TemporalGraphEdge> InEdges = new(new TemporalGraphEdgeComparer());

        /// <summary>
        /// Contains all edges to the out-neighbor
        /// </summary>
        public readonly Dictionary<TemporalGraphVertex, SortedSet<TemporalGraphEdge>> OutNeighbors = new();

        public readonly SortedSet<TemporalGraphEdge> OutEdges = new(new TemporalGraphEdgeComparer());


        public TemporalGraphVertex(long id, TemporalGraph temporalGraph)
        {
            Id = id;
            _graph = temporalGraph;
        }

        internal void UpdateType()
        {
            if (_graph.NoUpdate)
                return;

            var oldType = Type;
            var newType = (OutEdges.Count, InEdges.Count) switch
            {
                (0, 0) => VertexType.Isolated,
                ( > 0, 0) => VertexType.Source,
                (0, > 0) => VertexType.Sink,

                // less than also for non-strict, as a vertex with in and out-edges at the same time could possibly be in the solution
                ( > 0, > 0) when InEdges.First().Time < OutEdges.First().Time => VertexType.Gateway,

                _ => VertexType.Regular,
            };

            if (Type != newType)
            {
                Type = newType;
                VertexTypeChanged?.Invoke(this, oldType, newType);
            }
        }

        internal void RemoveEdge(TemporalGraphEdge edge)
        {
            if (edge.From == this)
            {
                if (!OutEdges.Remove(edge))
                    throw new InvalidOperationException($"Edge is not in out-edges set ({edge.From.Id} -> {edge.To.Id} @ {edge.Time.Value})");

                var outEdges = OutNeighbors[edge.To];
                if (!outEdges.Remove(edge))
                    throw new InvalidOperationException($"Edge is not in edge-set of in-neighbor ({edge.From.Id} -> {edge.To.Id} @ {edge.Time.Value})");

                if (outEdges.Count == 0)
                {
                    OutNeighbors.Remove(edge.To);
                }
            }
            else if (edge.To == this)
            {
                if (!InEdges.Remove(edge))
                    throw new InvalidOperationException($"Edge is not in in-edges set ({edge.From.Id} -> {edge.To.Id} @ {edge.Time.Value})");

                var inEdges = InNeighbors[edge.From];
                if (!inEdges.Remove(edge))
                    throw new InvalidOperationException($"Edge is not in edge-set of out-neighbor ({edge.From.Id} -> {edge.To.Id} @ {edge.Time.Value})");

                if (inEdges.Count == 0)
                {
                    InNeighbors.Remove(edge.From);
                }
            }
            else
            {
                throw new InvalidOperationException($"Edge does not connects to this vertex ({edge.From.Id} -> {edge.To.Id} @ {edge.Time.Value})");
            }

            UpdateType();
        }

        /// <summary>
        /// Adds an edge from this vertex to the specified vertex to both vertices.
        /// </summary>
        /// <param name="to"></param>
        /// <param name="time"></param>
        internal bool AddEdge(TemporalGraphEdge edge)
        {
            bool result;
            if (edge.From == this)
            {
                result = AddOutEdge(edge);
            }
            else if (edge.To == this)
            {
                result = AddInEdge(edge);
            }
            else
            {
                throw new InvalidOperationException();
            }

            UpdateType();
            return result;
        }

        private bool AddInEdge(TemporalGraphEdge edge)
        {
            if (!InNeighbors.TryGetValue(edge.From, out var inList))
            {
                inList = new SortedSet<TemporalGraphEdge>(new TemporalGraphEdgeComparer());
                InNeighbors.Add(edge.From, inList);
            }
            return InEdges.Add(edge) & inList.Add(edge);
        }

        private bool AddOutEdge(TemporalGraphEdge edge)
        {
            if (!OutNeighbors.TryGetValue(edge.To, out var outList))
            {
                outList = new SortedSet<TemporalGraphEdge>(new TemporalGraphEdgeComparer());
                OutNeighbors.Add(edge.To, outList);
            }

            return OutEdges.Add(edge) & outList.Add(edge);
        }

        public override int GetHashCode()
        {
            return (int)Id;
        }
    }
}
