﻿using System;
using System.Diagnostics;
using System.Threading;

namespace TemporalSourceCover
{
    [DebuggerDisplay("{Value}")]
    public readonly struct Time : IComparable<Time>
    {
        public static TimeMode Mode { get; set; }

        /// <summary>
        /// Returns whether an edge at timestep <paramref name="before"/> can proceed an edge at timestep <paramref name="after"/>.<br/>
        /// Strict:     before &lt; after<br/>
        /// Non-Strict: before &lt;= after
        /// </summary>
        /// <param name="before"></param>
        /// <param name="after"></param>
        /// <returns></returns>
        public static bool Proceeds(Time before, Time after)
        {
            switch (Mode)
            {
                case TimeMode.Strict: return before.Value < after.Value;
                case TimeMode.NonStrict: return before.Value <= after.Value;
                default: throw new InvalidOperationException();
            }
        }



        public static Time MinValue => new Time(long.MinValue);
        public static Time MaxValue => new Time(long.MaxValue);


        public readonly long Value;

        public Time(long value)
        {
            Value = value;
        }

        public static bool operator >(Time a, Time b) => a.Value > b.Value;
        public static bool operator >=(Time a, Time b) => a.Value >= b.Value;
        public static bool operator <(Time a, Time b) => a.Value < b.Value;
        public static bool operator <=(Time a, Time b) => a.Value <= b.Value;
        public static bool operator ==(Time a, Time b) => a.Value == b.Value;
        public static bool operator !=(Time a, Time b) => !(a == b);

        public static Time Max(Time a, Time b) => a > b ? a : b;
        public static Time Min(Time a, Time b) => a < b ? a : b;

        public int CompareTo(Time other)
        {
            return Value.CompareTo(other.Value);
        }

        public override bool Equals(object? obj)
        {
            return obj is Time time
                && time == this;
        }

        public override int GetHashCode()
        {
            return Value.GetHashCode();
        }
    }
}
