﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using TemporalSourceCover.DatasetTransformation;
using TemporalSourceCover.Utility;
using Timer = System.Threading.Timer;

namespace TemporalSourceCover
{
    /****************************************************/
    /************ This file is a mess. Sorry ************/
    /****************************************************/

    class Program
    {
        public static TimeSpan StepTimeout = TimeSpan.FromHours(1);
        public static TimeSpan HeuristicTimeout = TimeSpan.FromMinutes(30);

        static async Task Main(string[] args)
        {
            var now = DateTime.Now;
            var fileName = $"output-{now:yyyy-MM-dd_HH-mm-ss}";
            var dir = Directory.GetCurrentDirectory();

            if (args.Length == 1 && args[0].StartsWith("in:"))
            {
                var argsFile = args[0]["in:".Length..];
                Console.WriteLine($"Reading args from file: {argsFile}");
                args = ReadArgsFromFile(argsFile);

                fileName = $"output-{Path.GetFileNameWithoutExtension(argsFile)}-{now:yyyy-MM-dd_HH-mm-ss}";

                Console.WriteLine($"Arguments:\n{string.Join(" ", args)}");
            }
            // filter args with #
            args = args.Where(arg => !arg.StartsWith("#")).ToArray();

            // write to console and to file simultaneuosly
            var logFile = Path.Combine(dir, fileName);
#if DEBUG
            Console.SetOut(new ConsoleFileWriter(Console.Out, $"{logFile}.log"));
#endif

            if (args.Length < 2)
            {
                Console.WriteLine("No args found");
                return;
            }

            InitTimeout();
            InitHeuristicTimeout();
            Console.WriteLine($"Step timeout is {StepTimeout.TotalHours:###00}:{StepTimeout:mm\\:ss}");
            Console.WriteLine($"Heuristic timeout is {HeuristicTimeout.TotalHours:###00}:{HeuristicTimeout:mm\\:ss}");

            InitAlgorithmFeatures();
            Console.WriteLine($"Algorithm features:\n" +
                $"    Reduction Rules: {TemporalSourceCover.UseDataReductionRules}\n" +
                $"    Search-Tree Algorithm: {TemporalSourceCover.UseSearchTree}");

            InitReductionRuleConfig();
            Console.WriteLine($"Reduction rule config:\n" +
                $"    Temporal: {DataReductionRules.Config.ApplyTemporalRules}\n" +
                $"    Reachability: {DataReductionRules.Config.ApplyReachabilityRules}");


            // write current memory usage into console every 15s
            var memoryTimer = new Timer(_ =>
            {
                var process = Process.GetCurrentProcess();
                var bytes = process.WorkingSet64;
                var gb = bytes / (1024.0 * 1024.0 * 1024.0);
                Console.WriteLine($"[Memory Usage] {gb:0.000} GB");
            }, null, TimeSpan.Zero, TimeSpan.FromSeconds(15));


            CsvOutput.OutputPath = $"{logFile}.csv";
            CsvOutput.Heuristics.OutputPath = $"{logFile}-heuristics.csv";

            for (int i = 0; i < args.Length;)
            {
                bool IsNextArg(string arg)
                {
                    if (args.Length <= i || args[i] != arg)
                        return false;
                    i++;
                    return true;
                }

                var file = args[i++];
                var transformer = GetDatasetTransformer(args[i++]);
                var bidirectional = IsNextArg("--bidirectional");
                var heuristic = IsNextArg("--heuristic");

                string name = Path.GetFileNameWithoutExtension(file);

                foreach (TimeMode timeMode in new[] { TimeMode.Strict, TimeMode.NonStrict })
                {
                    // collect garbage in case the last graph used a lot of memory that has not been cleaned up yet
                    GC.Collect();

                    Time.Mode = timeMode;

                    Console.WriteLine($"Current mode: {timeMode}");
                    Console.WriteLine($"Directed: {!bidirectional}");
                    Console.WriteLine();

                    if (heuristic)
                    {
                        await FindHeuristicSolution(file, transformer, bidirectional, name, timeMode);
                    }
                    else
                    {
                        await FindSolution(file, transformer, bidirectional, name, timeMode);
                    }
                }
            }
        }

        private static Dictionary<string, Func<TemporalGraph, Solution>> _heuristics = new()
        {
            ["EarliestOutEdge"] = TemporalSourceCoverHeuristicSolutions.EarliestOutEdge,
            ["HighestNumberOfNeighbors"] = TemporalSourceCoverHeuristicSolutions.LargestOutNeighborhood,
            ["MaximumUnreachedOutNeighbors"] = TemporalSourceCoverHeuristicSolutions.LargestUnreachedOutNeighborhood,
        };
        private static async Task FindHeuristicSolution(string file, IDatasetTransformer transformer, bool bidirectional, string name, TimeMode timeMode)
        {
            using var scope = new CsvOutput.Heuristics.Scope();
            CsvOutput.Heuristics.CurrentScope.Name = name;
            CsvOutput.Heuristics.CurrentScope.Mode = timeMode;
            CsvOutput.Heuristics.CurrentScope.Directed = !bidirectional;

            try
            {
                Timeout.SetTimeout(HeuristicTimeout);

                TemporalGraph temporalGraph;
                using (var metrics = Metrics.BeginMeasurement($"Init graph \"{file}\""))
                {
                    temporalGraph = await GetTemporalGraphAsync(file, transformer, bidirectional);
                }

                foreach (var (heuristicName, heuristic) in _heuristics)
                {
                    GC.Collect();

                    var graph = temporalGraph.Clone();

                    Timeout.SetTimeout(HeuristicTimeout);
                    try
                    {
                        Solution solution;
                        using (var metrics = Metrics.BeginMeasurement($"Heuristic: {heuristicName}"))
                        {
                            solution = heuristic.Invoke(graph);
                            Console.WriteLine($"Solution size: {solution.Count}");

                            CsvOutput.Heuristics.CurrentScope.Stats.Add(new(heuristicName, metrics.Time, solution.Count));
                        }

#if VERIFY
                        VerifySolution(solution, temporalGraph);
#endif
                    }
                    catch (TaskCanceledException)
                    {
                        CsvOutput.Heuristics.CurrentScope.Stats.Add(new(heuristicName));
                    }
                }

                scope.Ok = true;
            }
            catch (TaskCanceledException)
            {
                Console.WriteLine("ABORTED Graph init");
            }
        }

        private static async Task FindSolution(string file, IDatasetTransformer transformer, bool bidirectional, string name, TimeMode timeMode)
        {
            using var scope = new CsvOutput.Scope();
            CsvOutput.CurrentScope.Name = name;
            CsvOutput.CurrentScope.Mode = timeMode;
            CsvOutput.CurrentScope.Directed = !bidirectional;

            try
            {
                Timeout.SetTimeout(StepTimeout);

                TemporalGraph temporalGraph;
                using (var metrics = Metrics.BeginMeasurement($"Init graph \"{file}\""))
                {
                    temporalGraph = await GetTemporalGraphAsync(file, transformer, bidirectional);

                    CsvOutput.CurrentScope.ReadInputTime = metrics.Time;
                }

                CsvOutput.CurrentScope.Lifetime = temporalGraph.Lifetime;

                TryProcessGraph(temporalGraph);
            }
            catch (TaskCanceledException)
            {
                Console.WriteLine("ABORTED Graph init");
            }
        }

        private static void InitTimeout()
        {
            var exePath = Process.GetCurrentProcess().MainModule?.FileName;
            if (exePath is null)
                return;

            var dir = Path.GetDirectoryName(exePath) ?? throw new InvalidOperationException();
            var timeoutFile = Path.Combine(dir, ".timeout");
            if (!File.Exists(timeoutFile))
                return;

            var content = File.ReadAllText(timeoutFile);
            var match = Regex.Match(content, @"^(?<hours>\d+):(?<minutes>\d+):(?<seconds>\d+)$");
            if (match.Success)
            {
                int hours = int.Parse(match.Groups["hours"].Value);
                int minutes = int.Parse(match.Groups["minutes"].Value);
                int seconds = int.Parse(match.Groups["seconds"].Value);

                StepTimeout = new TimeSpan(hours, minutes, seconds);
            }
        }

        private static void InitHeuristicTimeout()
        {
            var exePath = Process.GetCurrentProcess().MainModule?.FileName;
            if (exePath is null)
                return;

            var dir = Path.GetDirectoryName(exePath) ?? throw new InvalidOperationException();
            var timeoutFile = Path.Combine(dir, ".heuristic-timeout");
            if (!File.Exists(timeoutFile))
                return;

            var content = File.ReadAllText(timeoutFile);
            var match = Regex.Match(content, @"^(?<hours>\d+):(?<minutes>\d+):(?<seconds>\d+)$");
            if (match.Success)
            {
                int hours = int.Parse(match.Groups["hours"].Value);
                int minutes = int.Parse(match.Groups["minutes"].Value);
                int seconds = int.Parse(match.Groups["seconds"].Value);

                HeuristicTimeout = new TimeSpan(hours, minutes, seconds);
            }
        }

        private static void InitAlgorithmFeatures()
        {
            var exePath = Process.GetCurrentProcess().MainModule?.FileName;
            if (exePath is null)
                return;

            var dir = Path.GetDirectoryName(exePath) ?? throw new InvalidOperationException();
            var timeoutFile = Path.Combine(dir, ".algofeat");
            if (!File.Exists(timeoutFile))
                return;

            var content = File.ReadAllText(timeoutFile);
            TemporalSourceCover.UseDataReductionRules = content.Contains("rrules", StringComparison.OrdinalIgnoreCase);
            TemporalSourceCover.UseSearchTree = content.Contains("searchtree", StringComparison.OrdinalIgnoreCase);
        }

        private static void InitReductionRuleConfig()
        {
            var exePath = Process.GetCurrentProcess().MainModule?.FileName;
            if (exePath is null)
                return;

            var dir = Path.GetDirectoryName(exePath) ?? throw new InvalidOperationException();
            var timeoutFile = Path.Combine(dir, ".rrules");
            if (!File.Exists(timeoutFile))
                return;

            var content = File.ReadAllText(timeoutFile);
            DataReductionRules.Config.ApplyReachabilityRules = content.Contains("reachability", StringComparison.OrdinalIgnoreCase);
            DataReductionRules.Config.ApplyTemporalRules = content.Contains("temporal", StringComparison.OrdinalIgnoreCase);
        }

        private static void TryProcessGraph(TemporalGraph temporalGraph)
        {
            try
            {
                ProcessGraph(temporalGraph);
            }
            catch (Exception ex)
            {
                CsvOutput.CurrentScope.Ok = false;

                Console.WriteLine("AN ERROR OCCURED:");
                Console.WriteLine(ex);
            }
        }

        private static void PrintGraphStats(TemporalGraph temporalGraph)
        {
            Console.WriteLine($"Stats:");
            Console.WriteLine($"    #Vertices:      {temporalGraph.Vertices.Count}");
            Console.WriteLine($"        #Source:    {temporalGraph.Vertices.Values.Count(v => v.Type == VertexType.Source)}");
            Console.WriteLine($"        #Sink:      {temporalGraph.Vertices.Values.Count(v => v.Type == VertexType.Sink)}");
            Console.WriteLine($"        #Gateway:   {temporalGraph.Vertices.Values.Count(v => v.Type == VertexType.Gateway)}");
            Console.WriteLine($"        #Isolated:  {temporalGraph.Vertices.Values.Count(v => v.Type == VertexType.Isolated)}");
            Console.WriteLine($"        #Normal:    {temporalGraph.Vertices.Values.Count(v => v.Type == VertexType.Regular)}");
            Console.WriteLine();
            Console.WriteLine($"        #Red:       {temporalGraph.Vertices.Values.Count(v => v.IsRed)}");
            Console.WriteLine($"        #Black:     {temporalGraph.Vertices.Values.Count(v => !v.IsRed)}");
            Console.WriteLine($"    #Edges:         {temporalGraph.Edges.Count}");
        }

        private static void ProcessGraph(TemporalGraph temporalGraph)
        {
            // collect garbage in case the last graph used a lot of memory that has not been cleaned up yet
            GC.Collect();

            // initial graph stats
            CsvOutput.CurrentScope.Initial.VertexCount = temporalGraph.Vertices.Count;
            CsvOutput.CurrentScope.Initial.EdgeCount = temporalGraph.Edges.Count;
            foreach (var (type, vertices) in temporalGraph.VerticesByType)
            {
                CsvOutput.CurrentScope.Initial.CountByVertexType[type] = vertices.Count;
            }

            PrintGraphStats(temporalGraph);

#if VERIFY
            Timeout.SetTimeout(StepTimeout);
            Solution? gurobiSolution = null;
            try
            {
                var gurobiInstance = new TemporalSourceCoverInstance(temporalGraph, int.MaxValue, new HashSet<TemporalGraphVertex>());
                gurobiSolution = TemporalSourceCover.FindGurobiSolution(gurobiInstance, CsvOutput.CurrentScope.InitialGurobiStats, timeout: StepTimeout);

                VerifySolution(gurobiSolution, temporalGraph);
                Console.WriteLine($"Solution size: {gurobiSolution.Count}");
            }
            catch (TaskCanceledException)
            {
                Console.WriteLine("ABORTED initial Gurobi step");
            }
            // collect garbage in case
            GC.Collect();


#endif
            var verificationGraph = temporalGraph.Clone();

            Timeout.SetTimeout(StepTimeout);
            Solution partialSolutionAfterInitialDataReduction = new Solution();
            var instance = new TemporalSourceCoverInstance(temporalGraph, int.MaxValue, new HashSet<TemporalGraphVertex>());
            if (TemporalSourceCover.UseDataReductionRules)
            {
                try
                {
                    List<DataReductionResult> reductionResults;
                    using (var metrics = Metrics.BeginMeasurement("Apply Data Reduction Rules initially"))
                    {
                        partialSolutionAfterInitialDataReduction = DataReductionRules.Apply(instance, out reductionResults);

                        CsvOutput.CurrentScope.DataReductionTime = metrics.Time;
                    }

                    foreach (var reductionResult in reductionResults)
                    {
                        var data = CsvOutput.CurrentScope.DataReduction[reductionResult.Rule];
                        data.RemovedEdges += reductionResult.RemovedEdges;
                        data.RemovedVertices += reductionResult.RemovedVertices;
                        data.Time += reductionResult.Time;
                        data.TimesApplied += reductionResult.WasApplied ? 1 : 0;
                        data.VerticesAddedToSolution += reductionResult.SolutionVertices;
                    }

                    // collect garbage in case
                    GC.Collect();

                    Console.WriteLine($"In solution after reduction rules: {partialSolutionAfterInitialDataReduction.Count}");

                    // graph stats after initial data reduction
                    CsvOutput.CurrentScope.AfterDataReduction.VerticesInSolution = partialSolutionAfterInitialDataReduction.Count;
                    CsvOutput.CurrentScope.AfterDataReduction.VertexCount = temporalGraph.Vertices.Count;
                    CsvOutput.CurrentScope.AfterDataReduction.EdgeCount = temporalGraph.Edges.Count;
                    foreach (var (type, vertices) in temporalGraph.VerticesByType)
                    {
                        CsvOutput.CurrentScope.AfterDataReduction.CountByVertexType[type] = vertices.Count;
                    }
                    CsvOutput.CurrentScope.AfterDataReduction.BlackVertices = temporalGraph.BlackVertices.Count;
                    CsvOutput.CurrentScope.AfterDataReduction.RedVertices = temporalGraph.RedVertices.Count;

                    PrintDataReductionStats(reductionResults);
                    PrintGraphStats(temporalGraph);

                    var possibleSourceCount = instance.PossibleSources.Count();
                    Console.WriteLine($"Sources to check: {possibleSourceCount}");
                    CsvOutput.CurrentScope.PossibleSourcesAfterDataReduction = possibleSourceCount;

#if VERIFY
                    Timeout.SetTimeout(StepTimeout + StepTimeout);
                    VerifyReductionRules(gurobiSolution, partialSolutionAfterInitialDataReduction, instance, verificationGraph);
#else
                    if(possibleSourceCount == 0)
                    {
                        VerifySolution(partialSolutionAfterInitialDataReduction, verificationGraph);
                    }
#endif
                }
                catch (TaskCanceledException)
                {
                    Console.WriteLine("ABORTED initial data reduction or Gurobi verification");
                }
            }

            if (TemporalSourceCover.UseSearchTree)
            {
                Solution searchTreeSolution = new Solution();
                var minimalSolution = new Solution();
                Timeout.SetTimeout(StepTimeout);
                try
                {
                    CsvOutput.CurrentScope.SearchTreeAfterDataReduction.UseReductionRulesTemporal = DataReductionRules.Config.ApplyTemporalRules;
                    CsvOutput.CurrentScope.SearchTreeAfterDataReduction.UseReductionRulesReachability = DataReductionRules.Config.ApplyReachabilityRules;

                    using (var metrics = Metrics.BeginMeasurement("Find solution using search tree"))
                    {
                        searchTreeSolution = TemporalSourceCover.FindSolutionSearchTree(instance, out var calls);
                        CsvOutput.CurrentScope.SearchTreeAfterDataReduction.RunningTime = metrics.Time;
                        CsvOutput.CurrentScope.SearchTreeAfterDataReduction.Calls = calls;

                        minimalSolution.AddRange(searchTreeSolution);
                        minimalSolution.AddRange(partialSolutionAfterInitialDataReduction);

                        CsvOutput.CurrentScope.SearchTreeAfterDataReduction.SolutionSize = minimalSolution.Count;
                    }
                }
                catch (TaskCanceledException)
                {
                    Console.WriteLine("Search tree ABORTED");
                    return;
                }

#if VERIFY
                Console.WriteLine($"Found solution of {minimalSolution.Count} vertices. Verifying...");

                if (gurobiSolution is not null && minimalSolution.Count > gurobiSolution.Count)
                    throw new InvalidOperationException();

                VerifySolution(minimalSolution, verificationGraph);

                // break here if we found a solution smaller than Gurobis
                if (gurobiSolution is not null && minimalSolution.Count < gurobiSolution.Count)
                {
                    Console.WriteLine($"Found smaller solution than initial Gurobi ({minimalSolution.Count} vs. {gurobiSolution.Count})");
                }

                Console.WriteLine("Solution found by search tree is valid!");
#else
                Console.WriteLine($"Found solution of {minimalSolution.Count} vertices.");
#endif
            }

            Console.WriteLine();
            Console.WriteLine("-------------------------------------------------------");
            Console.WriteLine();
        }

        private static void PrintDataReductionStats(List<DataReductionResult> reductionResults)
        {
            Console.WriteLine("Data Reduction Stats:");
            if (reductionResults.Count == 0)
            {
                Console.WriteLine("    None");
                return;
            }

            foreach (var ruleResults in reductionResults.GroupBy(r => r.Rule))
            {
                var ruleResult = new DataReductionResult(ruleResults.Key);
                int timesApplied = 0;
                int timesCalled = 0;
                foreach (var r in ruleResults)
                {
                    ruleResult.Time += r.Time;
                    timesCalled++;

                    if (!r.WasApplied)
                        continue;

                    ruleResult.RemovedEdges += r.RemovedEdges;
                    ruleResult.RemovedVertices += r.RemovedVertices;
                    ruleResult.SolutionVertices += r.SolutionVertices;
                    ruleResult.IgnoredVertices += r.IgnoredVertices;
                    timesApplied++;
                }

                Console.WriteLine($"    {ruleResult.Rule} (applied {timesApplied} of {timesCalled} calls):");
                Console.WriteLine($"           Removed Vertices:     {ruleResult.RemovedVertices}");
                Console.WriteLine($"           Removed Edges:        {ruleResult.RemovedEdges}");
                Console.WriteLine($"           Added to Solution:    {ruleResult.SolutionVertices}");
                Console.WriteLine($"           Ignored for Solution: {ruleResult.IgnoredVertices}");
                Console.WriteLine($"           Time in s:            {ruleResult.Time.TotalSeconds}");
            }
        }

#if VERIFY
        private static void VerifyReductionRules(Solution? gurobiSolution, Solution solution, TemporalSourceCoverInstance instance, TemporalGraph verificationGraph)
        {
            Console.WriteLine($"Verifying reduction rules... {(gurobiSolution is object ? "With" : "Without")} initial Gurobi solution");
            Solution newGurobiSolution;
            try
            {
                newGurobiSolution = TemporalSourceCover.FindGurobiSolution(instance, CsvOutput.CurrentScope.GurobiStatsAfterDataReduction, epsilon: 1e-9, timeout: StepTimeout);
            }
            catch (TaskCanceledException)
            {
                Console.WriteLine("Gorubi verification cancelled");
                return;
            }

            newGurobiSolution.AddRange(solution);
            CsvOutput.CurrentScope.GurobiStatsAfterDataReduction.SolutionSize = newGurobiSolution.Count;

            if (gurobiSolution?.Count < newGurobiSolution.Count)
                throw new InvalidSolutionException(solution, null);

            VerifySolution(newGurobiSolution, verificationGraph);

            if (gurobiSolution?.Count > newGurobiSolution.Count)
            {
                Console.WriteLine($"Found solution smaller than Gurobis dominating set solution ({newGurobiSolution.Count} vs. {gurobiSolution.Count})");
            }
        }
#endif

        private static void VerifySolution(Solution solution, TemporalGraph verificationGraph)
        {
            Timeout.Disable();
            var graph = verificationGraph.Clone();
            var solutionVertices = solution.Select(v => graph.Vertices[v.Id]).ToList();

            var reachedVertices = new HashSet<TemporalGraphVertex>();
            foreach (var vertex in solutionVertices)
            {
                graph.RemoveSourceAndPaths(vertex, out var reachability, out _);
                reachedVertices.AddRange(reachability);

                if (reachedVertices.Count == verificationGraph.Vertices.Count)
                    return;
            }

            var notReached = verificationGraph.Vertices.Values.Where(n => !reachedVertices.Contains(n)).ToList();
            if (notReached.Count > 0)
            {
                throw new InvalidSolutionException(solution, notReached);
            }
        }

        private static async Task<TemporalGraph> GetTemporalGraphAsync(string file, IDatasetTransformer transformer, bool bidirectional)
        {
            var temporalGraph = new TemporalGraph();

            using var reader = new StreamReader(file);
            using (temporalGraph.DontUpdate())
            {
                await foreach (var edge in transformer.Transform(reader))
                {
                    Timeout.ThrowOnCancel();
                    temporalGraph.AddEdge(edge.From, edge.To, new Time(edge.Time), bidirectional);
                }
            }

            return temporalGraph;
        }


        private static IDatasetTransformer GetDatasetTransformer(string name)
        {
            var assembly = typeof(TUVTransformer).Assembly;
            var type = assembly.GetTypes().SingleOrDefault(type => type.Name == name)
                ?? throw new TypeLoadException($"The type \"{name}\" could not be found");

            if (!type.GetInterfaces().Contains(typeof(IDatasetTransformer)))
                throw new TypeLoadException($"The type \"{type.FullName}\" does not implement \"{typeof(IDatasetTransformer).FullName}\".");

            return (IDatasetTransformer)Activator.CreateInstance(type)!;
        }

        private static string[] ReadArgsFromFile(string file)
        {
            string input = File.ReadAllText(file);

            string dir = Path.GetDirectoryName(file)!;
            Console.WriteLine($"Current dir: '{dir}'");
            Directory.SetCurrentDirectory(dir);

            bool inQuote = false;
            return input.Split(c =>
            {
                if (c == '"')
                {
                    inQuote ^= true;
                }

                return !inQuote && char.IsWhiteSpace(c);
            }).Select(arg => TrimQuotes(arg.Trim()))
                .Where(arg => !string.IsNullOrEmpty(arg))
                .ToArray();


            static string TrimQuotes(string str)
            {
                if (str.Length >= 2 && str[0] == '"' && str[^1] == '"')
                    return str[1..^2];

                return str;
            }
        }
    }
}
